@@ -1,11 +1,11 @@
 /***************************************
-  $Header: /home/amb/cxref/RCS/cxref.c 1.35 1997/11/20 19:57:39 amb Exp $
+  $Header: /home/amb/cxref/RCS/cxref.c 1.52 1999/07/21 18:26:21 amb Exp $
 
-  C Cross Referencing & Documentation tool. Version 1.4a.
+  C Cross Referencing & Documentation tool. Version 1.5b.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1995,96,97 Andrew M. Bishop
+  This file Copyright 1995,96,97,98,99 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -14,6 +14,8 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+
+#include <limits.h>
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <sys/stat.h>
@@ -37,7 +39,7 @@
 
 static void Usage(int verbose);
 static int ParseConfigFile(void);
-static int ParseOptions(int nargs,char **args,int is_argv);
+static int ParseOptions(int nargs,char **args,int fromfile);
 
 static int DocumentTheFile(char* name);
 static FILE* popen_execvp(char** command);
@@ -45,6 +47,7 @@ static int pclose_execvp(FILE* f);
 
 static char** cpp_command;              /*+ The actual cpp command that is built up, adding -D, -U and -I options. +*/
 static int cpp_command_num=0;           /*+ The number of arguments to the cpp command. +*/
+static int cpp_argument_num=0;          /*+ The number of arguments to the -CPP argument. +*/
 
 /*+ The command line switch that sets the format of the output, +*/
 int option_all_comments=0,              /*+ use all comments. +*/
@@ -55,23 +58,35 @@ int option_all_comments=0,              /*+ use all comments. +*/
     option_warn=0,                      /*+ produce warnings. +*/
     option_index=0,                     /*+ produce an index. +*/
     option_raw=0,                       /*+ produce raw output. +*/
-    option_latex=0,                     /*+ produce latex output. +*/
-    option_html=0;                      /*+ produce html output. +*/
+    option_latex=0,                     /*+ produce LaTeX output. +*/
+    option_html=0,                      /*+ produce HTML output. +*/
+    option_rtf=0,                       /*+ produce RTF output. +*/
+    option_sgml=0;                      /*+ produce SGML output. +*/
 
 /*+ The option to control the mode of operation. +*/
-int option_delete=0;
+static int option_delete=0;
 
 /*+ The command line switch for the output name, +*/
-char *option_odir=".",                  /*+ the directory to use. +*/
-     *option_name="cxref",              /*+ the base part of the name. +*/
+char *option_odir=NULL,                 /*+ the directory to use. +*/
+     *option_name=NULL,                 /*+ the base part of the name. +*/
      *option_root=NULL;                 /*+ the source tree root directory. +*/
 
 /*+ The name of the include directories specified on the command line. +*/
 char **option_incdirs=NULL;
 
+/*+ The information about the cxref run, +*/
+char *run_command=NULL,         /*+ the command line options. +*/
+     *run_cpp_command=NULL;     /*+ the cpp command and options. +*/
+
 /*+ The number of include directories on the command line. +*/
 int option_nincdirs=0;
 
+/*+ The names of the files to process. +*/
+static char **option_files=NULL;
+
+/*+ The number of files to process. +*/
+static int option_nfiles=0;
+
 /*+ The current file that is being processed. +*/
 File CurFile=NULL;
 
@@ -90,31 +105,43 @@ int main(int argc,char** argv)
 {
  int i;
  char *root_prefix=NULL;
+ char here[PATH_MAX+1],there[PATH_MAX+1];
 
  if(argc==1)
     Usage(1);
 
+ /* Setup the variables. */
+
  cpp_command=(char**)Malloc(8*sizeof(char*));
  cpp_command[cpp_command_num++]=MallocString(CPP_COMMAND);
 
- for(i=1;i<strlen(CPP_COMMAND);i++)
-    if(cpp_command[0][i]==' ')
-       cpp_command[0][i]=0;
-    else
-       if(cpp_command[0][i-1]==0)
-         {
-          if((cpp_command_num%8)==6)
-             cpp_command=(char**)Realloc(cpp_command,(cpp_command_num+10)*sizeof(char*));
-          cpp_command[cpp_command_num++]=&cpp_command[0][i];
-         }
+ for(i=1;cpp_command[cpp_command_num-1][i];i++)
+    if(cpp_command[cpp_command_num-1][i]==' ')
+      {
+       cpp_command[cpp_command_num-1][i]=0;
+       if((cpp_command_num%8)==6)
+          cpp_command=(char**)Realloc(cpp_command,(cpp_command_num+10)*sizeof(char*));
+       cpp_command[cpp_command_num++]=MallocString(&cpp_command[cpp_command_num-1][i+1]);
+       i=1;
+      }
+
+ cpp_argument_num=cpp_command_num;
 
  option_incdirs=(char**)Malloc(8*sizeof(char*));
- option_incdirs[0]=".";
+ option_incdirs[0]=MallocString(".");
  option_nincdirs=1;
 
+ option_odir=MallocString(".");
+
+ option_name=MallocString("cxref");
+
+ option_files=(char**)Malloc(8*sizeof(char*));
+
+ run_command=argv[0];
+
  /* Parse the command line options. */
 
- if(ParseOptions(argc-1,&argv[1],1))
+ if(ParseOptions(argc-1,&argv[1],0))
     Usage(0);
 
  /* Parse the options in .cxref in this directory. */
@@ -126,46 +153,99 @@ int main(int argc,char** argv)
 
  if(option_root)
    {
-    char *here=(char*)Malloc(512),*there=(char*)Malloc(512);
-
-    if(!getcwd(there,255))
+    if(!getcwd(there,PATH_MAX))
       {fprintf(stderr,"cxref: Error cannot get current working directory (1).\n");exit(1);}
     if(chdir(option_root))
       {fprintf(stderr,"cxref: Error cannot change directory to '%s'.\n",option_root);exit(1);}
-    if(!getcwd(here,255))
-      {fprintf(stderr,"cxref: Error cannot get current working directory (2).\n");exit(1);}
+   }
 
-    if(!strncmp(here,there,strlen(here)))
+ if(!getcwd(here,PATH_MAX))
+   {fprintf(stderr,"cxref: Error cannot get current working directory (2).\n");exit(1);}
+
+ if(option_root)
+   {
+    if(!strcmp(here,there))
+       root_prefix=".";
+    else if(!strncmp(here,there,strlen(here)))
        root_prefix=there+strlen(here)+1;
     else
       {fprintf(stderr,"cxref: Error the -R option has not specified a parent directory of the current one.\n");exit(1);}
+   }
 
-    for(i=1;i<cpp_command_num;i++)
-       if(cpp_command[i][0]=='-' && cpp_command[i][1]=='I')
+ /* Modify the -I options for the new root directory. */
+
+ for(i=1;i<cpp_command_num;i++)
+    if(cpp_command[i][0]=='-' && cpp_command[i][1]=='I')
+      {
+       if(cpp_command[i][2]==0)
          {
-          if(cpp_command[i][2]==0)
-            {
-             if(cpp_command[++i][0]!='/')
-                cpp_command[i]=MallocString(CanonicaliseName(ConcatStrings(3,root_prefix,"/",cpp_command[i])));
-            }
-          else if(cpp_command[i][2]!='/')
-             cpp_command[i]=MallocString(ConcatStrings(2,"-I",CanonicaliseName(ConcatStrings(3,root_prefix,"/",cpp_command[i]+2))));
+          char *old=cpp_command[++i];
+          if(cpp_command[i][0]!='/' && root_prefix)
+             cpp_command[i]=MallocString(CanonicaliseName(ConcatStrings(3,root_prefix,"/",cpp_command[i])));
+          else if(cpp_command[i][0]=='/' && !strcmp(cpp_command[i],here))
+             cpp_command[i]=MallocString(".");
+          else if(cpp_command[i][0]=='/' && !strncmp(cpp_command[i],here,strlen(here)))
+             cpp_command[i]=MallocString(cpp_command[i]+strlen(here)+1);
+          else
+             cpp_command[i]=MallocString(CanonicaliseName(cpp_command[i]));
+          Free(old);
          }
+       else
+         {
+          char *old=cpp_command[i];
+          if(cpp_command[i][2]!='/' && root_prefix)
+             cpp_command[i]=MallocString(ConcatStrings(2,"-I",CanonicaliseName(ConcatStrings(3,root_prefix,"/",cpp_command[i]+2))));
+          else if(cpp_command[i][2]=='/' && !strcmp(&cpp_command[i][2],here))
+             cpp_command[i]=MallocString("-I.");
+          else if(cpp_command[i][2]=='/' && !strncmp(&cpp_command[i][2],here,strlen(here)))
+             cpp_command[i]=MallocString(ConcatStrings(2,"-I",cpp_command[i]+2+strlen(here)+1));
+          else
+             cpp_command[i]=MallocString(ConcatStrings(2,"-I",CanonicaliseName(cpp_command[i]+2)));
+          Free(old);
+         }
+      }
 
-    for(i=0;i<option_nincdirs;i++)
-       if(*option_incdirs[i]!='/')
-          option_incdirs[i]=MallocString(CanonicaliseName(ConcatStrings(3,root_prefix,"/",option_incdirs[i])));
+ for(i=0;i<option_nincdirs;i++)
+   {
+    char *old=option_incdirs[i];
+    if(*option_incdirs[i]!='/' && root_prefix)
+       option_incdirs[i]=MallocString(CanonicaliseName(ConcatStrings(3,root_prefix,"/",option_incdirs[i])));
+    else if(*option_incdirs[i]=='/' && !strcmp(option_incdirs[i],here))
+       option_incdirs[i]=MallocString(".");
+    else if(*option_incdirs[i]=='/' && !strncmp(option_incdirs[i],here,strlen(here)))
+       option_incdirs[i]=MallocString(option_incdirs[i]+strlen(here)+1);
+    else
+       option_incdirs[i]=MallocString(CanonicaliseName(option_incdirs[i]));
+    Free(old);
+   }
 
+ /* Parse the options in .cxref in the root directory. */
+
+ if(option_root)
     if(ParseConfigFile())
        Usage(0);
-   }
+
+ run_command=MallocString(run_command);
+
+ run_cpp_command=cpp_command[0];
+ for(i=1;i<cpp_command_num;i++)
+    run_cpp_command=ConcatStrings(3,run_cpp_command," ",cpp_command[i]);
+
+ run_cpp_command=MallocString(run_cpp_command);
+
+ TidyMemory();
+
+ /* Check the options for validity */
+
+ if(option_warn&WARN_XREF && !option_xref)
+    fprintf(stderr,"cxref: Warning using '-warn-xref' without '-xref'.\n");
 
  /* Process each file. */
 
- for(i=1;i<argc;i++)
-    if(argv[i])
+ if(option_files)
+    for(i=0;i<option_nfiles;i++)
       {
-       char *filename=CanonicaliseName(root_prefix?ConcatStrings(3,root_prefix,"/",argv[i]):argv[i]);
+       char *filename=CanonicaliseName(root_prefix?ConcatStrings(3,root_prefix,"/",option_files[i]):option_files[i]);
 
        if(!option_delete)
          {
@@ -174,7 +254,7 @@ int main(int argc,char** argv)
           if(!DocumentTheFile(filename))
             {
              if(option_xref)
-                CrossReference(CurFile);
+                CrossReference(CurFile,option_warn||option_raw||option_latex||option_html||option_rtf||option_sgml);
 
              if(option_raw || option_warn)
                 WriteWarnRawFile(CurFile);
@@ -182,6 +262,10 @@ int main(int argc,char** argv)
                 WriteLatexFile(CurFile);
              if(option_html)
                 WriteHTMLFile(CurFile);
+             if(option_rtf)
+                WriteRTFFile(CurFile);
+             if(option_sgml)
+                WriteSGMLFile(CurFile);
             }
 
           ResetLexer();
@@ -191,6 +275,8 @@ int main(int argc,char** argv)
           ResetVariableAnalyser();
           ResetFunctionAnalyser();
 
+          DeleteComment();
+
           DeleteFile(CurFile);
           CurFile=NULL;
          }
@@ -200,6 +286,8 @@ int main(int argc,char** argv)
 
           WriteLatexFileDelete(filename);
           WriteHTMLFileDelete(filename);
+          WriteRTFFileDelete(filename);
+          WriteSGMLFileDelete(filename);
          }
 
        TidyMemory();
@@ -225,6 +313,10 @@ int main(int argc,char** argv)
        WriteLatexAppendix(files,funcs,vars,types);
     if(option_html)
        WriteHTMLAppendix(files,funcs,vars,types);
+    if(option_rtf)
+       WriteRTFAppendix(files,funcs,vars,types);
+    if(option_sgml)
+       WriteSGMLAppendix(files,funcs,vars,types);
 
     DeleteStringList(files);
     DeleteStringList2(funcs);
@@ -236,11 +328,26 @@ int main(int argc,char** argv)
 
  /* Tidy up */
 
- Free(cpp_command[0]);
+ Free(option_odir);
+ Free(option_name);
+ if(option_root)
+    Free(option_root);
+
+ for(i=0;i<cpp_command_num;i++)
+    Free(cpp_command[i]);
  Free(cpp_command);
 
+ for(i=0;i<option_nincdirs;i++)
+    Free(option_incdirs[i]);
  Free(option_incdirs);
 
+ for(i=0;i<option_nfiles;i++)
+    Free(option_files[i]);
+ Free(option_files);
+
+ Free(run_command);
+ Free(run_cpp_command);
+
  PrintMemoryStatistics();
 
  return(0);
@@ -256,11 +363,11 @@ int main(int argc,char** argv)
 static void Usage(int verbose)
 {
  fputs("\n"
-       "              C Cross Referencing & Documenting tool - Version 1.4a\n"
+       "              C Cross Referencing & Documenting tool - Version 1.5a\n"
        "              -----------------------------------------------------\n"
        "\n"
-       "(c) Andrew M. Bishop 1995,96,97  [       amb@gedanken.demon.co.uk ]\n"
-       "                                 [http://www.gedanken.demon.co.uk/]\n"
+       "(c) Andrew M. Bishop 1995,96,97,98,99 [       amb@gedanken.demon.co.uk ]\n"
+       "                                      [http://www.gedanken.demon.co.uk/]\n"
        "\n"
        "Usage: cxref filename [ ... filename]\n"
        "             [-Odirname] [-Nbasename] [-Rdirname]\n"
@@ -269,7 +376,7 @@ static void Usage(int verbose)
        "             [-xref[-all][-file][-func][-var][-type]]\n"
        "             [-warn[-all][-comment][-xref]]\n"
        "             [-index[-all][-file][-func][-var][-type]]\n"
-       "             [-raw] [-latex|-latex2e] [-html]\n"
+       "             [-latex209|-latex2e] [-html20|-html32] [-rtf] [-sgml] [-raw]\n"
        "             [-Idirname] [-Ddefine] [-Udefine]\n"
        "             [-CPP cpp_program] [-- cpp_arg [ ... cpp_arg]]\n"
        "\n"
@@ -296,9 +403,11 @@ static void Usage(int verbose)
           "\n"
           "-index[-*]             : Produce a cross reference index (of specified types).\n"
           "\n"
-          "-raw                   : Produce raw output.\n"
-          "-latex | -latex2e      : Produce LaTeX output (version 2.09 or 2e).\n"
-          "-html                  : Produce HTML output (version 2.0).\n"
+          "-latex209 | -latex2e   : Produce LaTeX output (version 2.09 or 2e - default=2e).\n"
+          "-html20 | -html32      : Produce HTML output (version 2.0 or 3.2 - default=3.2).\n"
+          "-rtf                   : Produce RTF output (version 1.x).\n"
+          "-sgml                  : Produce SGML output (for SGML tools version 1.0.x).\n"
+          "-raw                   : Produce raw output .\n"
           "\n"
           "-I*, -D*, -U*          : The usual compiler switches.\n"
           "-CPP cpp_program       : The cpp program to use.\n"
@@ -319,7 +428,7 @@ static void Usage(int verbose)
 /*++++++++++++++++++++++++++++++++++++++
   Read in the options from the configuration file.
 
-  int ParseConfigFile Returns the values returned by ParseOptions().
+  int ParseConfigFile Returns the value returned by ParseOptions().
   ++++++++++++++++++++++++++++++++++++++*/
 
 static int ParseConfigFile(void)
@@ -327,11 +436,11 @@ static int ParseConfigFile(void)
  FILE *file=fopen(CXREF_CONFIG_FILE,"r");
  char **lines=NULL;
  int nlines=0;
- char data[129];
+ char data[257];
 
  if(file)
    {
-    while(fgets(data,80,file))
+    while(fgets(data,256,file))
       {
        char *d=data+strlen(data)-1;
 
@@ -356,8 +465,8 @@ static int ParseConfigFile(void)
           int i=2;
           while(data[i]==' ' || data[i]=='\t')
              data[i++]=0;
-          lines[nlines++]=MallocString(data);
-          lines[nlines++]=MallocString(data+i);
+          lines[nlines++]=CopyString(data);
+          lines[nlines++]=CopyString(data+i);
          }
        else if(!strncmp(data,"-CPP",4) &&
                (data[4]==' ' || data[4]=='\t'))
@@ -365,20 +474,32 @@ static int ParseConfigFile(void)
           int i=4;
           while(data[i]==' ' || data[i]=='\t')
              data[i++]=0;
-          lines[nlines++]=MallocString(data);
-          lines[nlines++]=MallocString(data+i);
+          lines[nlines++]=CopyString(data);
+          lines[nlines++]=CopyString(data+i);
          }
        else
           if(*data)
-             lines[nlines++]=MallocString(data);
+             lines[nlines++]=CopyString(data);
       }
 
     if(nlines)
       {
-       if(ParseOptions(nlines,lines,0))
+       int n_files=option_nfiles;
+
+       if(ParseOptions(nlines,lines,1))
+         {
+          fprintf(stderr,"cxref: Error parsing the .cxref file\n");
           return(1);
+         }
 
        Free(lines);
+
+       if(n_files!=option_nfiles)
+         {
+          for(;n_files<option_nfiles;n_files++)
+             fprintf(stderr,"cxref: File names '%s' only allowed on command line.\n",option_files[n_files]);
+          return(1);
+         }
       }
 
     fclose(file);
@@ -387,6 +508,7 @@ static int ParseConfigFile(void)
  return(0);
 }
 
+
 /*++++++++++++++++++++++++++++++++++++++
   Parse the options from the command line or from the .cxref file.
 
@@ -396,13 +518,12 @@ static int ParseConfigFile(void)
 
   char **args The actual arguments
 
-  int is_argv Set to true if these are the argc,argv values.
+  int fromfile A flag indicating that they are read from the .cxref file.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static int ParseOptions(int nargs,char **args,int is_argv)
+static int ParseOptions(int nargs,char **args,int fromfile)
 {
  int i,end_of_args=0;
- int cpp_original_num=cpp_command_num;
 
  for(i=0;i<nargs;i++)
    {
@@ -410,8 +531,8 @@ static int ParseOptions(int nargs,char **args,int is_argv)
       {
        if((cpp_command_num%8)==6)
           cpp_command=(char**)Realloc(cpp_command,(cpp_command_num+10)*sizeof(char*));
-       cpp_command[cpp_command_num++]=args[i];
-       if(is_argv) args[i]=NULL;
+       cpp_command[cpp_command_num++]=MallocString(args[i]);
+       run_command=ConcatStrings(3,run_command," ",args[i]);
        continue;
       }
 
@@ -420,135 +541,141 @@ static int ParseOptions(int nargs,char **args,int is_argv)
        char *incdir=NULL;
        if((cpp_command_num%8)==6)
           cpp_command=(char**)Realloc(cpp_command,(cpp_command_num+10)*sizeof(char*));
-       cpp_command[cpp_command_num++]=args[i];
+       cpp_command[cpp_command_num++]=MallocString(args[i]);
        if(args[i][2]==0)
          {
           if(args[i][1]=='I')
              incdir=args[i+1];
           if(i==nargs-1)
             {fprintf(stderr,"cxref: The -%c option requires a following argument.\n",args[i][1]);return(1);}
-          if(is_argv) args[i]=NULL;
           if((cpp_command_num%8)==6)
              cpp_command=(char**)Realloc(cpp_command,(cpp_command_num+10)*sizeof(char*));
-          cpp_command[cpp_command_num++]=args[++i];
+          run_command=ConcatStrings(3,run_command," ",args[i]);
+          cpp_command[cpp_command_num++]=MallocString(args[++i]);
          }
        else
           if(args[i][1]=='I')
              incdir=&args[i][2];
-       if(is_argv) args[i]=NULL;
 
        if(incdir)
          {
           if((option_nincdirs%8)==0)
              option_incdirs=(char**)Realloc(option_incdirs,(option_nincdirs+8)*sizeof(char*));
-          option_incdirs[option_nincdirs++]=incdir;
+          option_incdirs[option_nincdirs++]=MallocString(incdir);
          }
+
+       run_command=ConcatStrings(3,run_command," ",args[i]);
        continue;
       }
 
     if(!strcmp(args[i],"-CPP"))
       {
        char **old=cpp_command,*command;
-       int j,old_num=cpp_command_num;
+       int j,old_com_num=cpp_command_num,old_arg_num=cpp_argument_num;
 
-       if(args[i][4]==0)
-         {
-          if(i==nargs-1)
-            {fprintf(stderr,"cxref: The -CPP option requires a following argument.\n");return(1);}
-          if(is_argv) args[i]=NULL;
-          command=args[++i];
-         }
-       else
-          command=&args[i][4];
+       if(i==nargs-1)
+         {fprintf(stderr,"cxref: The -CPP option requires a following argument.\n");return(1);}
+       command=args[++i];
 
        cpp_command_num=0;
        cpp_command=(char**)Malloc(8*sizeof(char*));
        cpp_command[cpp_command_num++]=MallocString(command);
 
-       for(j=1;j<strlen(command);j++)
-          if(cpp_command[0][j]==' ')
-             cpp_command[0][j]=0;
-          else
-             if(cpp_command[0][j-1]==0)
-               {
-                if((cpp_command_num%8)==6)
-                   cpp_command=(char**)Realloc(cpp_command,(cpp_command_num+10)*sizeof(char*));
-                cpp_command[cpp_command_num++]=&cpp_command[0][j];
-               }
+       for(j=1;cpp_command[cpp_command_num-1][j];j++)
+          if(cpp_command[cpp_command_num-1][j]==' ')
+            {
+             cpp_command[cpp_command_num-1][j]=0;
+             if((cpp_command_num%8)==6)
+                cpp_command=(char**)Realloc(cpp_command,(cpp_command_num+10)*sizeof(char*));
+             cpp_command[cpp_command_num++]=MallocString(&cpp_command[cpp_command_num-1][j+1]);
+             j=1;
+            }
 
-       for(j=cpp_original_num;j<old_num;j++)
+       cpp_argument_num=cpp_command_num;
+
+       for(j=old_arg_num;j<old_com_num;j++)
          {
           if((cpp_command_num%8)==6)
              cpp_command=(char**)Realloc(cpp_command,(cpp_command_num+10)*sizeof(char*));
           cpp_command[cpp_command_num++]=old[j];
          }
 
-       cpp_original_num=cpp_command_num;
-       Free(old[0]); Free(old);
+       for(j=0;j<old_arg_num;j++)
+          Free(old[j]);
+       Free(old);
 
-       if(is_argv) args[i]=NULL;
+       run_command=ConcatStrings(4,run_command,"-CPP \"",args[i],"\"");
        continue;
       }
 
     if(!strncmp(args[i],"-O",2))
       {
+       if(option_odir)
+          Free(option_odir);
        if(args[i][2]==0)
          {
           if(i==nargs-1)
             {fprintf(stderr,"cxref: The -O option requires a following argument.\n");return(1);}
-          if(is_argv) args[i]=NULL;
-          option_odir=args[++i];
+          run_command=ConcatStrings(3,run_command," ",args[i]);
+          option_odir=MallocString(args[++i]);
          }
        else
-          option_odir=&args[i][2];
-       if(is_argv) args[i]=NULL;
+          option_odir=MallocString(&args[i][2]);
+       run_command=ConcatStrings(3,run_command," ",args[i]);
        continue;
       }
 
     if(!strncmp(args[i],"-N",2))
       {
+       if(option_name)
+          Free(option_name);
        if(args[i][2]==0)
          {
           if(i==nargs-1)
             {fprintf(stderr,"cxref: The -N option requires a following argument.\n");return(1);}
-          if(is_argv) args[i]=NULL;
-          option_name=args[++i];
+          run_command=ConcatStrings(3,run_command," ",args[i]);
+          option_name=MallocString(args[++i]);
          }
        else
-          option_name=&args[i][2];
-       if(is_argv) args[i]=NULL;
+          option_name=MallocString(&args[i][2]);
+       run_command=ConcatStrings(3,run_command," ",args[i]);
        continue;
       }
 
     if(!strncmp(args[i],"-R",2))
       {
+       if(option_root)
+          Free(option_root);
        if(args[i][2]==0)
          {
           if(i==nargs-1)
             {fprintf(stderr,"cxref: The -R option requires a following argument.\n");return(1);}
-          if(is_argv) args[i]=NULL;
-          option_root=args[++i];
+          run_command=ConcatStrings(3,run_command," ",args[i]);
+          option_root=MallocString(args[++i]);
          }
        else
-          option_root=&args[i][2];
-       if(is_argv) args[i]=NULL;
+          option_root=MallocString(&args[i][2]);
+       if(*option_root=='.' && !*(option_root+1))
+          option_root=NULL;
+       run_command=ConcatStrings(3,run_command," ",args[i]);
        continue;
       }
 
     if(!strcmp(args[i],"-delete"))
-      {option_delete=1; if(is_argv)args[i]=NULL; continue;}
+      {if(fromfile) {fprintf(stderr,"cxref: The -delete option cannot be used in the .cxref file.\n");return(1);}
+       option_delete=1; run_command=ConcatStrings(3,run_command," ",args[i]); continue;}
 
     if(!strcmp(args[i],"-all-comments"))
-      {option_all_comments=1; if(is_argv)args[i]=NULL; continue;}
+      {option_all_comments=1; run_command=ConcatStrings(3,run_command," ",args[i]); continue;}
 
     if(!strcmp(args[i],"-verbatim-comments"))
-      {option_verbatim_comments=1; if(is_argv)args[i]=NULL; continue;}
+      {option_verbatim_comments=1; run_command=ConcatStrings(3,run_command," ",args[i]); continue;}
 
     if(!strcmp(args[i],"-block-comments"))
-      {option_block_comments=1; if(is_argv)args[i]=NULL; continue;}
+      {option_block_comments=1; run_command=ConcatStrings(3,run_command," ",args[i]); continue;}
 
     if(!strcmp(args[i],"-no-comments"))
-      {option_no_comments=1; if(is_argv)args[i]=NULL; continue;}
+      {option_no_comments=1; run_command=ConcatStrings(3,run_command," ",args[i]); continue;}
 
     if(!strncmp(args[i],"-xref",5))
       {
@@ -566,7 +693,8 @@ static int ParseOptions(int nargs,char **args,int is_argv)
              if(!strncmp(p,"-type",5)) {option_xref|=XREF_TYPE; p=&p[5]; continue;}
              break;
             }
-       if(is_argv) args[i]=NULL;
+
+       run_command=ConcatStrings(3,run_command," ",args[i]);
        continue;
       }
 
@@ -584,7 +712,8 @@ static int ParseOptions(int nargs,char **args,int is_argv)
              if(!strncmp(p,"-xref"   ,5)) {option_warn|=WARN_XREF   ; p=&p[5]; continue;}
              break;
             }
-       if(is_argv) args[i]=NULL;
+
+       run_command=ConcatStrings(3,run_command," ",args[i]);
        continue;
       }
 
@@ -604,33 +733,48 @@ static int ParseOptions(int nargs,char **args,int is_argv)
              if(!strncmp(p,"-type",5)) {option_index|=INDEX_TYPE; p=&p[5]; continue;}
              break;
             }
-       if(is_argv) args[i]=NULL;
+
+       run_command=ConcatStrings(3,run_command," ",args[i]);
        continue;
       }
 
     if(!strcmp(args[i],"-raw"))
-      {option_raw=1; if(is_argv)args[i]=NULL; continue;}
+      {option_raw=1; run_command=ConcatStrings(3,run_command," ",args[i]); continue;}
 
-    if(!strcmp(args[i],"-latex"))
-      {option_latex=1; if(is_argv)args[i]=NULL; continue;}
-    if(!strcmp(args[i],"-latex2e"))
-      {option_latex=2; if(is_argv)args[i]=NULL; continue;}
+    if(!strcmp(args[i],"-latex209"))
+      {option_latex=1; run_command=ConcatStrings(3,run_command," ",args[i]); continue;}
+    if(!strcmp(args[i],"-latex2e") || !strcmp(args[i],"-latex"))
+      {option_latex=2; run_command=ConcatStrings(3,run_command," ",args[i]); continue;}
 
-    if(!strcmp(args[i],"-html"))
-      {option_html=1; if(is_argv)args[i]=NULL; continue;}
+    if(!strncmp(args[i],"-html20",7))
+      {option_html=1; if(!strcmp(args[i]+7,"-src"))option_html+=16;
+       run_command=ConcatStrings(3,run_command," ",args[i]); continue;}
+    if(!strncmp(args[i],"-html32",7))
+      {option_html=2; if(!strcmp(args[i]+7,"-src"))option_html+=16;
+       run_command=ConcatStrings(3,run_command," ",args[i]); continue;}
+    if(!strncmp(args[i],"-html",5))
+      {option_html=2; if(!strcmp(args[i]+5,"-src"))option_html+=16;
+       run_command=ConcatStrings(3,run_command," ",args[i]); continue;}
+
+    if(!strcmp(args[i],"-rtf"))
+      {option_rtf=1; run_command=ConcatStrings(3,run_command," ",args[i]); continue;}
+
+    if(!strcmp(args[i],"-sgml"))
+      {option_sgml=1; run_command=ConcatStrings(3,run_command," ",args[i]); continue;}
 
     if(!strcmp(args[i],"--"))
-      {end_of_args=1; if(is_argv)args[i]=NULL; continue;}
+      {end_of_args=1; run_command=ConcatStrings(3,run_command," ",args[i]); continue;}
 
     if(args[i][0]=='-')
       {fprintf(stderr,"cxref: Unknown option '%s'.\n",args[i]);return(1);}
 
-    if(!is_argv)
+    if(fromfile)
       {fprintf(stderr,"cxref: File names '%s' only allowed on command line.\n",args[i]);return(1);}
-   }
 
- if(option_warn&WARN_XREF && !option_xref)
-    fprintf(stderr,"cxref: Warning using '-warn-xref' without '-xref'.\n");
+    if(option_files && (option_nfiles%8)==0)
+       option_files=(char**)Realloc(option_files,(option_nfiles+8)*sizeof(char*));
+    option_files[option_nfiles++]=MallocString(args[i]);
+   }
 
  return(0);
 }
@@ -646,26 +790,32 @@ static int ParseOptions(int nargs,char **args,int is_argv)
 
 char *CanonicaliseName(char *name)
 {
- char *match;
+ char *match,*name2;
 
- while((match=strstr(name,"/../")))
+ match=name;
+ while((match=strstr(match,"/./")) || !strncmp(match=name,"./",2))
    {
-    char *prev=match; match+=4;
-    while(prev>name && *--prev!='/');
-    if(prev!=name)prev++;
-    while((*prev++=*match++));
+    char *prev=match, *next=match+2;
+    while((*prev++=*next++));
    }
 
- while((match=strstr(name,"/./")) || (match=strstr(name,"./"))==name)
+ match=name;
+ while((match=strstr(match,"//")))
    {
-    char *prev=match; match+=(*match=='/')?3:2;
-    while((*prev++=*match++));
+    char *prev=match, *next=match+1;
+    while((*prev++=*next++));
    }
 
- while((match=strstr(name,"//")))
+ match=name2=name;
+ while((match=strstr(match,"/../")))
    {
-    char *prev=match; match+=2;
-    while((*prev++=*match++));
+    char *prev=match, *next=match+4;
+    if((prev-name2)==2 && !strncmp(name2,"../",3))
+      {name2+=3;match++;continue;}
+    while(prev>name2 && *--prev!='/');
+    match=prev;
+    if(*prev=='/')prev++;
+    while((*prev++=*next++));
    }
 
  match=&name[strlen(name)-2];
@@ -675,11 +825,27 @@ char *CanonicaliseName(char *name)
  match=&name[strlen(name)-3];
  if(match>=name && !strcmp(match,"/.."))
     if(match==name)
-       *match=0;
+       *++match=0;
     else
        while(match>name && *--match!='/')
           *match=0;
 
+#if 1 /* as used in cxref */
+
+ match=&name[strlen(name)-1];
+ if(match>name && !strcmp(match,"/"))
+    *match=0;
+
+ if(!*name)
+    *name='.',*(name+1)=0;
+
+#else /* as used in wwwoffle */
+
+ if(!*name || !strncmp(name,"../",3))
+    *name='/',*(name+1)=0;
+
+#endif
+
  return(name);
 }
 
@@ -698,6 +864,7 @@ static int DocumentTheFile(char* name)
 {
  struct stat stat_buf;
  int error1,error2;
+ static int first=1;
 
  if(stat(name,&stat_buf)==-1)
    {fprintf(stderr,"cxref: Cannot access the file '%s'\n",name);return(1);}
@@ -710,7 +877,9 @@ static int DocumentTheFile(char* name)
  if(!yyin)
    {fprintf(stderr,"cxref: Failed to start the cpp command\n");exit(1);}
 
- yyrestart(yyin);
+ if(!first)
+    yyrestart(yyin);
+ first=0;
 
 #if YYDEBUG
  yydebug=(YYDEBUG==3);
@@ -728,7 +897,7 @@ static int DocumentTheFile(char* name)
 
 
 /*+ The process id of the pre-processor. +*/
-static int popen_pid;
+static pid_t popen_pid;
 
 /*++++++++++++++++++++++++++++++++++++++
   A popen function that takes a list of arguments not a string.
@@ -779,10 +948,15 @@ static FILE* popen_execvp(char** command)
 
 static int pclose_execvp(FILE* f)
 {
- int ret;
+ int status,ret;
 
- waitpid(popen_pid,&ret,0);
+ waitpid(popen_pid,&status,0);
  fclose(f);
 
+ if(WIFEXITED(status))
+    ret=WEXITSTATUS(status);
+ else
+    ret=-1;
+
  return(ret);
 }
