@@ -1,13 +1,13 @@
 /***************************************
-  $Header: /home/amb/cxref/RCS/latex.c 1.27 1997/11/20 19:19:07 amb Exp $
+  $Header: /home/amb/cxref/RCS/latex.c 1.34 1999/05/15 16:58:57 amb Exp $
 
-  C Cross Referencing & Documentation tool. Version 1.4a.
+  C Cross Referencing & Documentation tool. Version 1.5a.
 
   Writes the Latex output.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1995,96,97 Andrew M. Bishop
+  This file Copyright 1995,96,97,98 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -47,6 +47,10 @@ extern char* option_odir;
 /*+ The base name of the file for the output. +*/
 extern char* option_name;
 
+/*+ The information about the cxref run, +*/
+extern char *run_command,       /*+ the command line options. +*/
+            *run_cpp_command;   /*+ the cpp command and options. +*/
+
 extern char *latex_fonts_style,*latex_page_style,*latex_cxref_style;
 
 static void WriteLatexFilePart(File file);
@@ -61,7 +65,7 @@ static void WriteLatexFunction(Function func,char* filename);
 static void WriteLatexDocument(char* name,int appendix);
 static void WriteLatexTemplate(char* name);
 
-static char* latex(char* c);
+static char* latex(char* c,int verbatim);
 
 /*+ The output file for the latex. +*/
 static FILE* of;
@@ -69,6 +73,7 @@ static FILE* of;
 /*+ Counts the lines in a table to insert breaks. +*/
 static int countlines=0;
 
+
 /*++++++++++++++++++++++++++++++++++++++
   Write a Latex file for a complete File structure and all components.
 
@@ -111,7 +116,10 @@ void WriteLatexFile(File file)
  /* Write out a header. */
 
  fputs("% This LaTeX file generated by cxref\n",of);
- fputs("% cxref program (c) Andrew M. Bishop 1995,96,97.\n",of);
+ fputs("% cxref program (c) Andrew M. Bishop 1995,96,97,98,99.\n",of);
+ fputs("\n",of);
+ fprintf(of,"%% Cxref: %s %s\n",run_command,file->name);
+ fprintf(of,"%% CPP  : %s\n",run_cpp_command);
  fputs("\n",of);
 
  /*+ The file structure is broken into its components and they are each written out. +*/
@@ -219,9 +227,9 @@ void WriteLatexFile(File file)
 
  fclose(of);
 
-/* Clear the memory in latex() */
+ /* Clear the memory in latex() */
 
- latex(NULL); latex(NULL); latex(NULL); latex(NULL);
+ latex(NULL,0); latex(NULL,0); latex(NULL,0); latex(NULL,0);
 }
 
 
@@ -235,13 +243,13 @@ static void WriteLatexFilePart(File file)
 {
  int i;
 
- fprintf(of,"\\markboth{File %s}{File %s}\n",latex(file->name),latex(file->name));
- fprintf(of,"\\section{File %s}\n",latex(file->name));
+ fprintf(of,"\\markboth{File %s}{File %s}\n",latex(file->name,0),latex(file->name,0));
+ fprintf(of,"\\section{File %s}\n",latex(file->name,0));
  fprintf(of,"\\label{file_%s}\n\n",file->name);
 
  if(file->comment)
     if(option_verbatim_comments)
-       fprintf(of,"\\begin{verbatim}\n%s\n\\end{verbatim}\n\n",latex(file->comment));
+       fprintf(of,"\\begin{verbatim}\n%s\n\\end{verbatim}\n\n",latex(file->comment,1));
     else
       {
        char *rcs1=strstr(file->comment,"$Header"),*rcs2=NULL;
@@ -251,15 +259,15 @@ static void WriteLatexFilePart(File file)
           if(rcs2)
             {
              rcs2[0]=0;
-             fprintf(of,"{\\bf RCS %s}\n\n",latex(&rcs1[1]));
+             fprintf(of,"{\\bf RCS %s}\n\n",latex(&rcs1[1],0));
              fprintf(of,"\\smallskip\n");
              rcs2[0]='$';
             }
          }
        if(rcs2)
-          fprintf(of,"%s\n\n",latex(&rcs2[2]));
+          fprintf(of,"%s\n\n",latex(&rcs2[2],0));
        else
-          fprintf(of,"%s\n\n",latex(file->comment));
+          fprintf(of,"%s\n\n",latex(file->comment,0));
       }
 
  if(file->inc_in->n)
@@ -270,10 +278,10 @@ static void WriteLatexFilePart(File file)
        fprintf(of,"\\medskip\n");
     fprintf(of,"\\begin{cxreftabii}\nIncluded in:");
     for(i=0;i<file->inc_in->n;i++)
-      {/* Allow a break in every 8 (or so) items to stop allow the table to break over the page. */
+      {/* Allow a break in every 8 (or so) items to allow the table to break over the page. */
        if(min(i,file->inc_in->n-i)%8 == 4)
           fprintf(of,"\\cxreftabbreak{cxreftabii}\n");
-       fprintf(of,"\\ & %s & \\cxreffile{%s}\\\\\n",latex(file->inc_in->s[i]),file->inc_in->s[i]);
+       fprintf(of,"\\ & %s & \\cxreffile{%s}\\\\\n",latex(file->inc_in->s[i],0),file->inc_in->s[i]);
       }
     fprintf(of,"\\end{cxreftabii}\n\n");
    }
@@ -287,28 +295,25 @@ static void WriteLatexFilePart(File file)
     if(file->f_refs->n)
       {
        int others=0;
+
+       fprintf(of,"Refs Func:");
+
        for(i=0;i<file->f_refs->n;i++)
-         {
           if(file->f_refs->s2[i])
             {
+             fprintf(of,"\\ & %s() & %s & \\cxreffunc{%s}{%s}\\\\\n",latex(file->f_refs->s1[i],0),latex(file->f_refs->s2[i],0),file->f_refs->s1[i],file->f_refs->s2[i]);
              if(++tabcount%8 == 4)
                 fprintf(of,"\\cxreftabbreak{cxreftabiii}\n");
-             if(i==others)
-                fprintf(of,"Refs Func:");
-             fprintf(of,"\\ & %s() & %s & \\cxreffunc{%s}{%s}\\\\\n",latex(file->f_refs->s1[i]),latex(file->f_refs->s2[i]),file->f_refs->s1[i],file->f_refs->s2[i]);
             }
           else
              others++;
-         }
 
        if(others)
          {
-          if(others==file->f_refs->n)
-             fprintf(of,"Refs Func:");
           fprintf(of,"\\ & \\cxreftabiiispan{");
           for(i=0;i<file->f_refs->n;i++)
              if(!file->f_refs->s2[i])
-                fprintf(of,--others?"%s(), ":"%s()",latex(file->f_refs->s1[i]));
+                fprintf(of,--others?"%s(), ":"%s()",latex(file->f_refs->s1[i],0));
           fprintf(of,"} &\\\\\n");
          }
       }
@@ -316,28 +321,25 @@ static void WriteLatexFilePart(File file)
     if(file->v_refs->n)
       {
        int others=0;
+
+       fprintf(of,"Refs Var:");
+
        for(i=0;i<file->v_refs->n;i++)
-         {
           if(file->v_refs->s2[i])
             {
+             fprintf(of,"\\ & %s & %s & \\cxrefvar{%s}{%s}\\\\\n",latex(file->v_refs->s1[i],0),latex(file->v_refs->s2[i],0),file->v_refs->s1[i],file->v_refs->s2[i]);
              if(++tabcount%8 == 4)
                 fprintf(of,"\\cxreftabbreak{cxreftabiii}\n");
-             if(i==others)
-                fprintf(of,"Refs Var:");
-             fprintf(of,"\\ & %s & %s & \\cxrefvar{%s}{%s}\\\\\n",latex(file->v_refs->s1[i]),latex(file->v_refs->s2[i]),file->v_refs->s1[i],file->v_refs->s2[i]);
             }
           else
              others++;
-         }
 
        if(others)
          {
-          if(others==file->v_refs->n)
-             fprintf(of,"Refs Var:");
           fprintf(of,"\\ & \\cxreftabiiispan{");
           for(i=0;i<file->v_refs->n;i++)
              if(!file->v_refs->s2[i])
-                fprintf(of,--others?" %s,":" %s",latex(file->v_refs->s1[i]));
+                fprintf(of,--others?" %s,":" %s",latex(file->v_refs->s1[i],0));
           fprintf(of,"} &\\\\\n");
          }
       }
@@ -356,14 +358,14 @@ static void WriteLatexFilePart(File file)
 static void WriteLatexInclude(Include inc)
 {
  if(inc->comment)
-    fprintf(of,"%s\n\n\\smallskip\n",latex(inc->comment));
+    fprintf(of,"%s\n\n\\smallskip\n",latex(inc->comment,0));
 
  fprintf(of,"\\begin{cxreftabi}\n"); countlines=1;
 
  if(inc->scope==LOCAL)
-    fprintf(of,"{\\stt \\#include \"%s\"} &\\cxreffile{%s}\\\\\n",latex(inc->name),inc->name);
+    fprintf(of,"{\\stt \\#include \"%s\"} &\\cxreffile{%s}\\\\\n",latex(inc->name,0),inc->name);
  else
-    fprintf(of,"{\\stt \\#include <%s>} &\\\\\n",latex(inc->name));
+    fprintf(of,"{\\stt \\#include <%s>} &\\\\\n",latex(inc->name,0));
 
  if(inc->includes)
     WriteLatexSubInclude(inc->includes,1);
@@ -390,9 +392,9 @@ static void WriteLatexSubInclude(Include inc,int depth)
     fprintf(of,"\\hspace*{%3.1fin}",0.2*depth);
 
     if(inc->scope==LOCAL)
-       fprintf(of,"{\\stt \\#include \"%s\"} &\\cxreffile{%s}\\\\\n",latex(inc->name),inc->name);
+       fprintf(of,"{\\stt \\#include \"%s\"} &\\cxreffile{%s}\\\\\n",latex(inc->name,0),inc->name);
     else
-       fprintf(of,"{\\stt \\#include <%s>} &\\\\\n",latex(inc->name));
+       fprintf(of,"{\\stt \\#include <%s>} &\\\\\n",latex(inc->name,0));
 
     if(inc->includes)
        WriteLatexSubInclude(inc->includes,depth+1);
@@ -414,18 +416,18 @@ static void WriteLatexDefine(Define def)
  int pargs=0;
 
  if(def->comment)
-    fprintf(of,"%s\n\n\\smallskip\n",latex(def->comment));
+    fprintf(of,"%s\n\n\\smallskip\n",latex(def->comment,0));
 
- fprintf(of,"{\\stt \\#define %s",latex(def->name));
+ fprintf(of,"{\\stt \\#define %s",latex(def->name,0));
 
  if(def->value)
-    fprintf(of," %s",latex(def->value));
+    fprintf(of," %s",latex(def->value,0));
 
  if(def->args->n)
    {
     fprintf(of,"( ");
     for(i=0;i<def->args->n;i++)
-       fprintf(of,i?", %s":"%s",latex(def->args->s1[i]));
+       fprintf(of,i?", %s":"%s",latex(def->args->s1[i],0));
     fprintf(of," )");
    }
  fprintf(of,"}\n\n");
@@ -439,7 +441,7 @@ static void WriteLatexDefine(Define def)
     fprintf(of,"\\smallskip\n");
     fprintf(of,"\\begin{cxrefarglist}\n");
     for(i=0;i<def->args->n;i++)
-       fprintf(of,"\\cxrefargitem{%s} %s\n",latex(def->args->s1[i]),def->args->s2[i]?latex(def->args->s2[i]):"\\ ");
+       fprintf(of,"\\cxrefargitem{%s} %s\n",latex(def->args->s1[i],0),def->args->s2[i]?latex(def->args->s2[i],0):"\\ ");
     fprintf(of,"\\end{cxrefarglist}\n\n");
    }
 }
@@ -456,9 +458,9 @@ static void WriteLatexDefine(Define def)
 static void WriteLatexTypedef(Typedef type,char* filename)
 {
  if(type->type)
-    fprintf(of,"\n\\subsubsection{Typedef %s}\n",latex(type->name));
+    fprintf(of,"\n\\subsubsection{Typedef %s}\n",latex(type->name,0));
  else
-    fprintf(of,"\n\\subsubsection{Type %s}\n",latex(type->name));
+    fprintf(of,"\n\\subsubsection{Type %s}\n",latex(type->name,0));
 
  if(!strncmp("enum",type->name,4))
     fprintf(of,"\\label{type_enum_%s_%s}\n\n",&type->name[5],filename);
@@ -472,10 +474,10 @@ static void WriteLatexTypedef(Typedef type,char* filename)
           fprintf(of,"\\label{type_%s_%s}\n\n",type->name,filename);
 
  if(type->comment)
-    fprintf(of,"%s\n\n\\smallskip\n",latex(type->comment));
+    fprintf(of,"%s\n\n\\smallskip\n",latex(type->comment,0));
 
  if(type->type)
-    fprintf(of,"{\\stt typedef %s}\n\n",latex(type->type));
+    fprintf(of,"{\\stt typedef %s}\n\n",latex(type->type,0));
 
  if(type->sutype)
    {
@@ -490,16 +492,16 @@ static void WriteLatexTypedef(Typedef type,char* filename)
        fprintf(of,"\\smallskip\n");
        fprintf(of,"\\begin{cxreftabii}\n");
        if(type->typexref->type)
-          fprintf(of,"See:& Typedef %s & \\cxreftype{%s}{%s}\\\\\n",latex(type->typexref->name),type->typexref->name,filename);
+          fprintf(of,"See:& Typedef %s & \\cxreftype{%s}{%s}\\\\\n",latex(type->typexref->name,0),type->typexref->name,filename);
        else
           if(!strncmp("enum",type->typexref->name,4))
-             fprintf(of,"See:& Type %s & \\cxreftype{enum_%s}{%s}\\\\\n",latex(type->typexref->name),&type->typexref->name[5],filename);
+             fprintf(of,"See:& Type %s & \\cxreftype{enum_%s}{%s}\\\\\n",latex(type->typexref->name,0),&type->typexref->name[5],filename);
           else
              if(!strncmp("union",type->typexref->name,5))
-                fprintf(of,"See:& Type %s & \\cxreftype{union_%s}{%s}\\\\\n",latex(type->typexref->name),&type->typexref->name[6],filename);
+                fprintf(of,"See:& Type %s & \\cxreftype{union_%s}{%s}\\\\\n",latex(type->typexref->name,0),&type->typexref->name[6],filename);
              else
                 if(!strncmp("struct",type->typexref->name,6))
-                   fprintf(of,"See:& Type %s & \\cxreftype{struct_%s}{%s}\\\\\n",latex(type->typexref->name),&type->typexref->name[7],filename);
+                   fprintf(of,"See:& Type %s & \\cxreftype{struct_%s}{%s}\\\\\n",latex(type->typexref->name,0),&type->typexref->name[7],filename);
        fprintf(of,"\\end{cxreftabii}\n\n");
       }
 }
@@ -526,11 +528,13 @@ static void WriteLatexStructUnion(StructUnion su, int depth)
  fprintf(of,"\\hspace*{%3.1fin}",0.2*depth);
 
  if(depth && su->comment && !su->comps)
-    fprintf(of,"{\\stt %s;} & %s \\\\\n",latex(su->name),latex(su->comment));
+    fprintf(of,"{\\stt %s;} & %s \\\\\n",latex(su->name,0),latex(su->comment,0));
+ else if(!depth || su->comps)
+    fprintf(of,"{\\stt %s} &\\\\\n",latex(su->name,0));
  else
-    fprintf(of,"{\\stt %s%s} &\\\\\n",latex(su->name),su->comps?"":";");
+    fprintf(of,"{\\stt %s;} &\\\\\n",latex(su->name,0));
 
- if(su->comps)
+ if(!depth || su->comps)
    {
     fprintf(of,"\\hspace*{%3.1fin}",0.1+0.2*depth);
     fprintf(of,"{\\stt \\{} &\\\\\n");
@@ -544,9 +548,9 @@ static void WriteLatexStructUnion(StructUnion su, int depth)
       {
        fprintf(of,"\\hspace*{%3.1fin}",0.1+0.2*depth);
        if(depth && su->comment)
-          fprintf(of,"{\\stt %s;} & %s \\\\\n",splitsu[5]?latex(&splitsu[6]):"",latex(su->comment));
+          fprintf(of,"{\\stt %s;} & %s \\\\\n",splitsu[5]?latex(&splitsu[6],0):"",latex(su->comment,0));
        else
-          fprintf(of,"{\\stt %s;} &\\\\\n",splitsu[5]?latex(&splitsu[6]):"");
+          fprintf(of,"{\\stt %s;} &\\\\\n",splitsu[5]?latex(&splitsu[6],0):"");
       }
    }
 
@@ -567,14 +571,14 @@ static void WriteLatexVariable(Variable var,char* filename)
  int i;
 
  if(var->scope&GLOBAL)
-    fprintf(of,"\n\\subsubsection{Variable %s}\n",latex(var->name));
+    fprintf(of,"\n\\subsubsection{Variable %s}\n",latex(var->name,0));
  else
-    fprintf(of,"{\\bf %s}\n",latex(var->name));
+    fprintf(of,"{\\bf %s}\n",latex(var->name,0));
 
  fprintf(of,"\\label{var_%s_%s}\n\n",var->name,filename);
 
  if(var->comment)
-    fprintf(of,"%s\n\n\\smallskip\n",latex(var->comment));
+    fprintf(of,"%s\n\n\\smallskip\n",latex(var->comment,0));
 
  fprintf(of,"{\\stt ");
 
@@ -584,7 +588,7 @@ static void WriteLatexVariable(Variable var,char* filename)
     if(!(var->scope&GLOBAL) && var->scope&(EXTERNAL|EXTERN_F))
        fprintf(of,"extern ");
 
- fprintf(of,"%s}\n\n",latex(var->type));
+ fprintf(of,"%s}\n\n",latex(var->type,0));
 
  if(var->scope&(GLOBAL|LOCAL))
    {
@@ -594,17 +598,17 @@ static void WriteLatexVariable(Variable var,char* filename)
        fprintf(of,"\\begin{cxreftabiii}\n");
 
        if(var->incfrom)
-          fprintf(of,"Inc. from:& %s & \\ & \\cxrefvar{%s}{%s}\\\\\n",latex(var->incfrom),var->name,var->incfrom);
+          fprintf(of,"Inc. from:& %s & \\ & \\cxrefvar{%s}{%s}\\\\\n",latex(var->incfrom,0),var->name,var->incfrom);
 
        for(i=0;i<var->visible->n;i++)
          {
           if(min(i,var->visible->n+var->used->n-i)%8 == 4)
              fprintf(of,"\\cxreftabbreak{cxreftabiii}\n");
           if(i==0) fprintf(of,"Visible in:");
-          if(var->visible->s1[i][0]=='$')
-             fprintf(of,"\\ & %s & \\ & \\cxreffile{%s}\\\\\n",latex(var->visible->s2[i]),var->visible->s2[i]);
+          if(var->visible->s1[i][0]=='$' && !var->visible->s1[i][1])
+             fprintf(of,"\\ & %s & \\ & \\cxreffile{%s}\\\\\n",latex(var->visible->s2[i],0),var->visible->s2[i]);
           else
-             fprintf(of,"\\ & %s() & %s & \\cxreffunc{%s}{%s}\\\\\n",latex(var->visible->s1[i]),latex(var->visible->s2[i]),var->visible->s1[i],var->visible->s2[i]);
+             fprintf(of,"\\ & %s() & %s & \\cxreffunc{%s}{%s}\\\\\n",latex(var->visible->s1[i],0),latex(var->visible->s2[i],0),var->visible->s1[i],var->visible->s2[i]);
          }
 
        for(i=0;i<var->used->n;i++)
@@ -612,13 +616,13 @@ static void WriteLatexVariable(Variable var,char* filename)
           if(min(i,var->visible->n+var->used->n-i)%8 == 4)
              fprintf(of,"\\cxreftabbreak{cxreftabiii}\n");
           if(i==0) fprintf(of,"Used in:");
-          if(var->used->s1[i][0]=='$')
-             fprintf(of,"\\ & %s & \\ & \\cxreffile{%s}\\\\\n",latex(var->used->s2[i]),var->used->s2[i]);
+          if(var->used->s1[i][0]=='$' && !var->used->s1[i][1])
+             fprintf(of,"\\ & %s & \\ & \\cxreffile{%s}\\\\\n",latex(var->used->s2[i],0),var->used->s2[i]);
           else
              if(var->scope&LOCAL)
-                fprintf(of,"\\ & %s() & \\ & \\cxreffunc{%s}{%s}\\\\\n",latex(var->used->s1[i]),var->used->s1[i],var->used->s2[i]);
+                fprintf(of,"\\ & %s() & \\ & \\cxreffunc{%s}{%s}\\\\\n",latex(var->used->s1[i],0),var->used->s1[i],var->used->s2[i]);
              else
-                fprintf(of,"\\ & %s() & %s & \\cxreffunc{%s}{%s}\\\\\n",latex(var->used->s1[i]),latex(var->used->s2[i]),var->used->s1[i],var->used->s2[i]);
+                fprintf(of,"\\ & %s() & %s & \\cxreffunc{%s}{%s}\\\\\n",latex(var->used->s1[i],0),latex(var->used->s2[i],0),var->used->s1[i],var->used->s2[i]);
          }
 
        fprintf(of,"\\end{cxreftabiii}\n\n");
@@ -629,7 +633,7 @@ static void WriteLatexVariable(Variable var,char* filename)
       {
        fprintf(of,"\\smallskip\n");
        fprintf(of,"\\begin{cxreftabiii}\n");
-       fprintf(of,"Defined in:& %s & \\ & \\cxrefvar{%s}{%s}\\\\\n",latex(var->defined),var->name,var->defined);
+       fprintf(of,"Defined in:& %s & \\ & \\cxrefvar{%s}{%s}\\\\\n",latex(var->defined,0),var->name,var->defined);
        fprintf(of,"\\end{cxreftabiii}\n\n");
       }
 }
@@ -649,20 +653,20 @@ static void WriteLatexFunction(Function func,char* filename)
  char* comment2=NULL,*type;
 
  if(func->scope&GLOBAL)
-    fprintf(of,"\n\\subsubsection{Global Function %s()}\n",latex(func->name));
+    fprintf(of,"\n\\subsubsection{Global Function %s()}\n",latex(func->name,0));
  else
-    fprintf(of,"\n\\subsubsection{Local Function %s()}\n",latex(func->name));
+    fprintf(of,"\n\\subsubsection{Local Function %s()}\n",latex(func->name,0));
  fprintf(of,"\\label{func_%s_%s}\n\n",func->name,filename);
 
  if(func->comment)
     if(option_verbatim_comments)
-       fprintf(of,"\\begin{verbatim}\n%s\n\\end{verbatim}\n\n",latex(func->comment));
+       fprintf(of,"\\begin{verbatim}\n%s\n\\end{verbatim}\n\n",latex(func->comment,1));
     else
       {
        comment2=strstr(func->comment,"\n\n");
        if(comment2)
           comment2[0]=0;
-       fprintf(of,"%s\n\n",latex(func->comment));
+       fprintf(of,"%s\n\n",latex(func->comment,0));
        fprintf(of,"\\smallskip\n");
       }
 
@@ -675,10 +679,10 @@ static void WriteLatexFunction(Function func,char* filename)
 
  if((type=strstr(func->type,"()")))
     type[0]=0;
- fprintf(of,"%s ( ",latex(func->type));
+ fprintf(of,"%s ( ",latex(func->type,0));
 
  for(i=0;i<func->args->n;i++)
-    fprintf(of,i?", %s":"%s",latex(func->args->s1[i]));
+    fprintf(of,i?", %s":"%s",latex(func->args->s1[i],0));
 
  if(type)
    {fprintf(of," %s}\n\n",&type[1]);type[0]='(';}
@@ -694,17 +698,17 @@ static void WriteLatexFunction(Function func,char* filename)
     fprintf(of,"\\smallskip\n");
     fprintf(of,"\\begin{cxrefarglist}\n");
     if(pret)
-       fprintf(of,"\\cxrefargitem{%s} %s\n",latex(func->type),func->cret?latex(func->cret):"\\ ");
+       fprintf(of,"\\cxrefargitem{%s} %s\n",latex(func->type,0),func->cret?latex(func->cret,0):"\\ ");
     if(pargs)
        for(i=0;i<func->args->n;i++)
-          fprintf(of,"\\cxrefargitem{%s} %s\n",latex(func->args->s1[i]),func->args->s2[i]?latex(func->args->s2[i]):"\\ ");
+          fprintf(of,"\\cxrefargitem{%s} %s\n",latex(func->args->s1[i],0),func->args->s2[i]?latex(func->args->s2[i],0):"\\ ");
     fprintf(of,"\\end{cxrefarglist}\n\n");
    }
 
  if(comment2)
    {
     fprintf(of,"\\smallskip\n");
-    fprintf(of,"%s\n\n",latex(&comment2[2]));
+    fprintf(of,"%s\n\n",latex(&comment2[2],0));
     comment2[0]='\n';
    }
 
@@ -715,92 +719,86 @@ static void WriteLatexFunction(Function func,char* filename)
     fprintf(of,"\\begin{cxreftabiii}\n");
 
     if(func->protofile)
-       fprintf(of,"Prototype:& %s & \\ & \\cxreffile{%s}\\\\\n",latex(func->protofile),func->protofile);
+       fprintf(of,"Prototype:& %s & \\ & \\cxreffile{%s}\\\\\n",latex(func->protofile,0),func->protofile);
 
     if(func->incfrom)
-       fprintf(of,"Inc. from:& %s & \\ & \\cxreffunc{%s}{%s}\\\\\n",latex(func->incfrom),func->name,func->incfrom);
+       fprintf(of,"Inc. from:& %s & \\ & \\cxreffunc{%s}{%s}\\\\\n",latex(func->incfrom,0),func->name,func->incfrom);
 
     if(func->calls->n)
       {
        int others=0;
+
+       fprintf(of,"Calls:");
+
        for(i=0;i<func->calls->n;i++)
-         {
           if(func->calls->s2[i])
             {
+             fprintf(of,"\\ & %s() & %s & \\cxreffunc{%s}{%s}\\\\\n",latex(func->calls->s1[i],0),latex(func->calls->s2[i],0),func->calls->s1[i],func->calls->s2[i]);
              if(++tabcount%8 == 4)
                 fprintf(of,"\\cxreftabbreak{cxreftabiii}\n");
-             if(i==others)
-                fprintf(of,"Calls:");
-             fprintf(of,"\\ & %s() & %s & \\cxreffunc{%s}{%s}\\\\\n",latex(func->calls->s1[i]),latex(func->calls->s2[i]),func->calls->s1[i],func->calls->s2[i]);
             }
           else
              others++;
-         }
 
        if(others)
          {
-          if(others==func->calls->n)
-             fprintf(of,"Calls:");
           fprintf(of,"\\ & \\cxreftabiiispan{");
           for(i=0;i<func->calls->n;i++)
              if(!func->calls->s2[i])
-                fprintf(of,--others?" %s(),":" %s()",latex(func->calls->s1[i]));
+                fprintf(of,--others?" %s(),":" %s()",latex(func->calls->s1[i],0));
           fprintf(of,"} &\\\\\n");
          }
       }
 
     if(func->called->n)
       {
+       fprintf(of,"Called by:");
+
        for(i=0;i<func->called->n;i++)
          {
+          fprintf(of,"\\ & %s() & %s & \\cxreffunc{%s}{%s}\\\\\n",latex(func->called->s1[i],0),latex(func->called->s2[i],0),func->called->s1[i],func->called->s2[i]);
           if(++tabcount%8 == 4)
              fprintf(of,"\\cxreftabbreak{cxreftabiii}\n");
-          if(i==0)
-             fprintf(of,"Called by:");
-          fprintf(of,"\\ & %s() & %s & \\cxreffunc{%s}{%s}\\\\\n",latex(func->called->s1[i]),latex(func->called->s2[i]),func->called->s1[i],func->called->s2[i]);
          }
       }
 
     if(func->used->n)
       {
+       fprintf(of,"Used in:");
+
        for(i=0;i<func->used->n;i++)
          {
+          if(func->used->s1[i][0]=='$' && !func->used->s1[i][1])
+             fprintf(of,"\\ & %s & \\ & \\cxreffile{%s}\\\\\n",latex(func->used->s2[i],0),func->used->s2[i]);
+          else
+             fprintf(of,"\\ & %s() & %s & \\cxreffunc{%s}{%s}\\\\\n",latex(func->used->s1[i],0),latex(func->used->s2[i],0),func->used->s1[i],func->used->s2[i]);
           if(++tabcount%8 == 4)
              fprintf(of,"\\cxreftabbreak{cxreftabiii}\n");
-          if(i==0)
-             fprintf(of,"Used in:");
-          if(func->used->s1[i][0]=='$')
-             fprintf(of,"\\ & %s & \\ & \\cxreffile{%s}\\\\\n",latex(func->used->s2[i]),func->used->s2[i]);
-          else
-             fprintf(of,"\\ & %s() & %s & \\cxreffunc{%s}{%s}\\\\\n",latex(func->used->s1[i]),latex(func->used->s2[i]),func->used->s1[i],func->used->s2[i]);
          }
       }
 
     if(func->f_refs->n)
       {
        int others=0;
+
+       fprintf(of,"Refs Func:");
+
        for(i=0;i<func->f_refs->n;i++)
-         {
           if(func->f_refs->s2[i])
             {
+             fprintf(of,"\\ & %s() & %s & \\cxreffunc{%s}{%s}\\\\\n",latex(func->f_refs->s1[i],0),latex(func->f_refs->s2[i],0),func->f_refs->s1[i],func->f_refs->s2[i]);
              if(++tabcount%8 == 4)
                 fprintf(of,"\\cxreftabbreak{cxreftabiii}\n");
-             if(i==others)
-                fprintf(of,"Refs Func:");
-             fprintf(of,"\\ & %s() & %s & \\cxreffunc{%s}{%s}\\\\\n",latex(func->f_refs->s1[i]),latex(func->f_refs->s2[i]),func->f_refs->s1[i],func->f_refs->s2[i]);
             }
           else
              others++;
-         }
 
        if(others)
          {
-          if(others==func->f_refs->n)
-             fprintf(of,"Refs Func:");
           fprintf(of,"\\ & \\cxreftabiiispan{");
           for(i=0;i<func->f_refs->n;i++)
              if(!func->f_refs->s2[i])
-                fprintf(of,--others?" %s(),":" %s()",latex(func->f_refs->s1[i]));
+                fprintf(of,--others?" %s(),":" %s()",latex(func->f_refs->s1[i],0));
           fprintf(of,"} &\\\\\n");
          }
       }
@@ -808,28 +806,25 @@ static void WriteLatexFunction(Function func,char* filename)
     if(func->v_refs->n)
       {
        int others=0;
+
+       fprintf(of,"Refs Var:");
+
        for(i=0;i<func->v_refs->n;i++)
-         {
           if(func->v_refs->s2[i])
             {
+             fprintf(of,"\\ & %s & %s & \\cxrefvar{%s}{%s}\\\\\n",latex(func->v_refs->s1[i],0),latex(func->v_refs->s2[i],0),func->v_refs->s1[i],func->v_refs->s2[i]);
              if(++tabcount%8 == 4)
                 fprintf(of,"\\cxreftabbreak{cxreftabiii}\n");
-             if(i==others)
-                fprintf(of,"Refs Var:");
-             fprintf(of,"\\ & %s & %s & \\cxrefvar{%s}{%s}\\\\\n",latex(func->v_refs->s1[i]),latex(func->v_refs->s2[i]),func->v_refs->s1[i],func->v_refs->s2[i]);
             }
           else
              others++;
-         }
 
        if(others)
          {
-          if(others==func->v_refs->n)
-             fprintf(of,"Refs Var:");
           fprintf(of,"\\ & \\cxreftabiiispan{");
           for(i=0;i<func->v_refs->n;i++)
              if(!func->v_refs->s2[i])
-                fprintf(of,--others?" %s,":" %s",latex(func->v_refs->s1[i]));
+                fprintf(of,--others?" %s,":" %s",latex(func->v_refs->s1[i],0));
           fprintf(of,"} &\\\\\n");
          }
       }
@@ -931,7 +926,7 @@ static void WriteLatexTemplate(char* name)
    {fprintf(stderr,"cxref: Failed to open the main LaTeX output file '%s'\n",name);exit(1);}
 
  fputs("% This LaTeX file generated by cxref\n",template);
- fputs("% cxref program (c) Andrew M. Bishop 1995,96,97.\n",template);
+ fputs("% cxref program (c) Andrew M. Bishop 1995,96,97,98,99.\n",template);
  fputs("\n",template);
  if(option_latex==1)
     fputs("\\documentstyle[fonts,page,cxref]{report}\n",template);
@@ -946,8 +941,8 @@ static void WriteLatexTemplate(char* name)
  fputs("\n",template);
  fputs("% Contents (Optional, either here or at end)\n",template);
  fputs("\n",template);
- fputs("\\markboth{Contents}{Contents}\n",template);
- fputs("\\tableofcontents\n",template);
+ fputs("%\\markboth{Contents}{Contents}\n",template);
+ fputs("%\\tableofcontents\n",template);
  fputs("\n",template);
  fputs("\\chapter{Source Files}\n",template);
  fputs("\n",template);
@@ -957,8 +952,8 @@ static void WriteLatexTemplate(char* name)
  fputs("\n",template);
  fputs("% Contents (Optional, either here or at beginning)\n",template);
  fputs("\n",template);
- fputs("%\\markboth{Contents}{Contents}\n",template);
- fputs("%\\tableofcontents\n",template);
+ fputs("\\markboth{Contents}{Contents}\n",template);
+ fputs("\\tableofcontents\n",template);
  fputs("\n",template);
  fputs("\\end{document}\n",template);
 
@@ -1026,6 +1021,15 @@ void WriteLatexAppendix(StringList files,StringList2 funcs,StringList2 vars,Stri
  if(!of)
    {fprintf(stderr,"cxref: Failed to open the LaTeX appendix file '%s'\n",ofile);exit(1);}
 
+ /* Write out a header. */
+
+ fputs("% This LaTeX file generated by cxref\n",of);
+ fputs("% cxref program (c) Andrew M. Bishop 1995,96,97,98,99.\n",of);
+ fputs("\n",of);
+ fprintf(of,"%% Cxref: %s\n",run_command);
+ fprintf(of,"%% CPP  : %s\n",run_cpp_command);
+ fputs("\n",of);
+
  /* Write the file structure out */
 
  fprintf(of,"\\chapter{Cross References}\n");
@@ -1041,7 +1045,7 @@ void WriteLatexAppendix(StringList files,StringList2 funcs,StringList2 vars,Stri
       {
        if(min(i,files->n-i)%8 == 4)
           fprintf(of,"\\cxreftabbreak{cxreftabiib}\n");
-       fprintf(of,"%s & \\ & \\cxreffile{%s}\\\\\n",latex(files->s[i]),files->s[i]);
+       fprintf(of,"%s & \\ & \\cxreffile{%s}\\\\\n",latex(files->s[i],0),files->s[i]);
       }
     fprintf(of,"\\end{cxreftabiib}\n\n");
    }
@@ -1057,7 +1061,7 @@ void WriteLatexAppendix(StringList files,StringList2 funcs,StringList2 vars,Stri
       {
        if(min(i,funcs->n-i)%8 == 4)
           fprintf(of,"\\cxreftabbreak{cxreftabiib}\n");
-       fprintf(of,"%s & %s & \\cxreffunc{%s}{%s}\\\\\n",latex(funcs->s1[i]),latex(funcs->s2[i]),funcs->s1[i],funcs->s2[i]);
+       fprintf(of,"%s & %s & \\cxreffunc{%s}{%s}\\\\\n",latex(funcs->s1[i],0),latex(funcs->s2[i],0),funcs->s1[i],funcs->s2[i]);
       }
     fprintf(of,"\\end{cxreftabiib}\n\n");
    }
@@ -1073,7 +1077,7 @@ void WriteLatexAppendix(StringList files,StringList2 funcs,StringList2 vars,Stri
       {
        if(min(i,vars->n-i)%8 == 4)
           fprintf(of,"\\cxreftabbreak{cxreftabiib}\n");
-       fprintf(of,"%s & %s & \\cxrefvar{%s}{%s}\\\\\n",latex(vars->s1[i]),latex(vars->s2[i]),vars->s1[i],vars->s2[i]);
+       fprintf(of,"%s & %s & \\cxrefvar{%s}{%s}\\\\\n",latex(vars->s1[i],0),latex(vars->s2[i],0),vars->s1[i],vars->s2[i]);
       }
     fprintf(of,"\\end{cxreftabiib}\n\n");
    }
@@ -1090,24 +1094,24 @@ void WriteLatexAppendix(StringList files,StringList2 funcs,StringList2 vars,Stri
        if(min(i,types->n-i)%8 == 4)
           fprintf(of,"\\cxreftabbreak{cxreftabiib}\n");
        if(!strncmp("enum",types->s1[i],4))
-          fprintf(of,"%s & %s & \\cxreftype{enum_%s}{%s}\\\\\n",latex(types->s1[i]),latex(types->s2[i]),&types->s1[i][5],types->s2[i]);
+          fprintf(of,"%s & %s & \\cxreftype{enum_%s}{%s}\\\\\n",latex(types->s1[i],0),latex(types->s2[i],0),&types->s1[i][5],types->s2[i]);
        else
           if(!strncmp("union",types->s1[i],5))
-             fprintf(of,"%s & %s & \\cxreftype{union_%s}{%s}\\\\\n",latex(types->s1[i]),latex(types->s2[i]),&types->s1[i][6],types->s2[i]);
+             fprintf(of,"%s & %s & \\cxreftype{union_%s}{%s}\\\\\n",latex(types->s1[i],0),latex(types->s2[i],0),&types->s1[i][6],types->s2[i]);
           else
              if(!strncmp("struct",types->s1[i],6))
-                fprintf(of,"%s & %s & \\cxreftype{struct_%s}{%s}\\\\\n",latex(types->s1[i]),latex(types->s2[i]),&types->s1[i][7],types->s2[i]);
+                fprintf(of,"%s & %s & \\cxreftype{struct_%s}{%s}\\\\\n",latex(types->s1[i],0),latex(types->s2[i],0),&types->s1[i][7],types->s2[i]);
              else
-                fprintf(of,"%s & %s & \\cxreftype{%s}{%s}\\\\\n",latex(types->s1[i]),latex(types->s2[i]),types->s1[i],types->s2[i]);
+                fprintf(of,"%s & %s & \\cxreftype{%s}{%s}\\\\\n",latex(types->s1[i],0),latex(types->s2[i],0),types->s1[i],types->s2[i]);
       }
     fprintf(of,"\\end{cxreftabiib}\n\n");
    }
 
  fclose(of);
 
-/* Clear the memory in latex() */
+ /* Clear the memory in latex(,0) */
 
- latex(NULL); latex(NULL); latex(NULL); latex(NULL);
+ latex(NULL,0); latex(NULL,0); latex(NULL,0); latex(NULL,0);
 }
 
 
@@ -1174,98 +1178,89 @@ void WriteLatexFileDelete(char *name)
 
   char* c A non-safe LaTeX string.
 
+  int verbatim Set to true inside a verbatim environment.
+
   The function can only be called four times in each fprintf() since it returns one of only four static strings.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static char* latex(char* c)
+static char* latex(char* c,int verbatim)
 {
  static char safe[4][256],*malloced[4]={NULL,NULL,NULL,NULL};
  static int which=0;
- int i=0,j=0,len=256-12;             /* 12 is the longest possible inserted amount */
- int eol=1,copy=0,skip=0;
+ int copy=0,skip=0;
+ int i=0,j=0,delta=13,len=256-delta;
  char* ret;
 
  which=(which+1)%4;
  ret=safe[which];
 
+ safe[which][0]=0;
+
  if(malloced[which])
    {Free(malloced[which]);malloced[which]=NULL;}
 
  if(c)
-    do
-      {
+   {
+    i=CopyOrSkip(c,"latex",&copy,&skip);
+
+    while(1)
+       {
        for(;j<len && c[i];i++)
+         {
           if(copy)
             {ret[j++]=c[i]; if(c[i]=='\n') copy=0;}
           else if(skip)
             {               if(c[i]=='\n') skip=0;}
-          else switch(c[i])
-            {
-            case '<':
-            case '>':
-             ret[j++]='$';
+          else if(verbatim)
              ret[j++]=c[i];
-             ret[j++]='$';
-             break;
-            case '\\':
-             strcpy(&ret[j],"$\\backslash$");j+=12;
-             break;
-            case '~':
-             strcpy(&ret[j],"$\\sim$");j+=6;
-             break;
-            case '^':
-             strcpy(&ret[j],"$\\wedge$");j+=8;
-             break;
-            case '#':
-            case '$':
-            case '%':
-            case '&':
-            case '_':
-            case '{':
-            case '}':
-             ret[j++]='\\';
-             ret[j++]=c[i];
-             break;
-            case '\n':
-             eol=1;
-             ret[j++]=c[i];
-             break;
-            default:
-             if(eol)
+          else
+             switch(c[i])
                {
-                if(!strncmp(c+i,"+latex+",7) || !strncmp(c+i,"-html-",6))
-                  {
-                   do { i++; } while(c[i]!='-' && c[i]!='+'); i++;
-                   copy=1;
-                   break;
-                  }
-                if(!strncmp(c+i,"-latex-",7) || !strncmp(c+i,"+html+",6) || !strncmp(c+i,"+none+",6))
-                  {
-                   do { i++; } while(c[i]!='-' && c[i]!='+'); i++;
-                   skip=1;
-                   break;
-                  }
-                if(c[i]!=' ' && c[i]!='\t')
-                   eol=0;
+               case '<':
+               case '>':
+                ret[j++]='$';
+                ret[j++]=c[i];
+                ret[j++]='$';
+                break;
+               case '\\':
+                strcpy(&ret[j],"$\\backslash$");j+=12;
+                break;
+               case '~':
+                strcpy(&ret[j],"$\\sim$");j+=6;
+                break;
+               case '^':
+                strcpy(&ret[j],"$\\wedge$");j+=8;
+                break;
+               case '#':
+               case '$':
+               case '%':
+               case '&':
+               case '_':
+               case '{':
+               case '}':
+                ret[j++]='\\';
+                ret[j++]=c[i];
+                break;
+               default:
+                ret[j++]=c[i];
                }
-             ret[j++]=c[i];
-            }
+          if(c[i]=='\n')
+             i+=CopyOrSkip(c+i,"latex",&copy,&skip);
+         }
 
        if(c[i])                 /* Not finished */
          {
           if(malloced[which])
-             malloced[which]=Realloc(malloced[which],len+256+12);
+             malloced[which]=Realloc(malloced[which],len+delta+256);
           else
-            {malloced[which]=Malloc(len+256+12); strncpy(malloced[which],ret,(unsigned)j);}
+            {malloced[which]=Malloc(len+delta+256); strncpy(malloced[which],ret,(unsigned)j);}
           ret=malloced[which];
           len+=256;
          }
        else
-         {ret[j]=0; ret=NULL;}
+         {ret[j]=0; break;}
       }
-    while(ret);
- else
-    safe[which][0]=0;
+   }
 
- return(malloced[which]?malloced[which]:safe[which]);
+ return(ret);
 }
