@@ -1,13 +1,13 @@
 /***************************************
-  $Header: /home/amb/cxref/RCS/warn-raw.c 1.23 1997/11/20 19:19:07 amb Exp $
+  $Header: /home/amb/cxref/RCS/warn-raw.c 1.27 1999/03/19 17:37:51 amb Exp $
 
-  C Cross Referencing & Documentation tool. Version 1.4a.
+  C Cross Referencing & Documentation tool. Version 1.5a.
 
   Writes the raw information and / or warnings out.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1995,96,97 Andrew M. Bishop
+  This file Copyright 1995,96,97,99 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -234,6 +234,9 @@ static void WriteWarnRawDefine(Define def)
  if(option_warn&WARN_COMMENT && !def->comment)
     printf("Warning %16s : #Define %s does not have a comment.\n",filename,def->name);
 
+ if(option_raw)
+    printf("Defined: %s:%d\n",filename,def->lineno);
+
  for(i=0;i<def->args->n;i++)
    {
     if(option_raw)
@@ -266,6 +269,9 @@ static void WriteWarnRawTypedef(Typedef type)
  if(option_warn&WARN_COMMENT && !type->comment)
     printf("Warning %16s : Type %s does not have a comment.\n",filename,type->name);
 
+ if(option_raw)
+    printf("Defined: %s:%d\n",filename,type->lineno);
+
  if(option_raw)
     if(type->type)
        printf("Type: %s\n",type->type);
@@ -305,11 +311,13 @@ static void WriteWarnRawStructUnion(StructUnion su, int depth,StructUnion base)
     for(i=0;i<depth;i++) printf("   ");
     if(depth && su->comment && !su->comps)
        printf("%s; <<<%s>>>\n",su->name,su->comment);
+    else if(!depth || su->comps)
+       printf("%s\n",su->name);
     else
-       printf("%s%s\n",su->name,su->comps?"":";");
+       printf("%s;\n",su->name);
    }
 
- if(su->comps)
+ if(!depth || su->comps)
    {
     if(option_raw)
       {
@@ -366,6 +374,9 @@ static void WriteWarnRawVariable(Variable var)
  if(option_warn&WARN_COMMENT && !var->comment && (var->scope&(GLOBAL|LOCAL|EXTERNAL|EXTERN_F) || option_raw))
     printf("Warning %16s : Variable %s does not have a comment.\n",filename,var->name);
 
+ if(option_raw)
+    printf("Defined: %s:%d\n",var->incfrom?var->incfrom:filename,var->lineno);
+
  if(option_raw)
     printf("Type: %s\n",var->type);
 
@@ -382,13 +393,13 @@ static void WriteWarnRawVariable(Variable var)
        if(var->scope&(GLOBAL|LOCAL))
          {
           for(i=0;i<var->visible->n;i++)
-             if(var->visible->s1[i][0]=='$')
+             if(var->visible->s1[i][0]=='$' && !var->visible->s1[i][1])
                 printf("Visible in %s\n",var->visible->s2[i]);
              else
                 printf("Visible in %s : %s\n",var->visible->s1[i],var->visible->s2[i]);
 
           for(i=0;i<var->used->n;i++)
-             if(var->used->s1[i][0]=='$')
+             if(var->used->s1[i][0]=='$' && !var->used->s1[i][1])
                 printf("Used in %s\n",var->used->s2[i]);
              else
                 printf("Used in %s : %s\n",var->used->s1[i],var->used->s2[i]);
@@ -449,6 +460,9 @@ static void WriteWarnRawFunction(Function func)
  if(option_warn&WARN_COMMENT && !func->comment)
     printf("Warning %16s : Function %s() does not have a comment.\n",filename,func->name);
 
+ if(option_raw)
+    printf("Defined: %s:%d\n",filename,func->lineno);
+
  if(option_xref&XREF_FUNC)
    {
     if(func->protofile && option_raw)
@@ -501,7 +515,7 @@ static void WriteWarnRawFunction(Function func)
     if(option_raw)
        for(i=0;i<func->used->n;i++)
          {
-          if(func->used->s1[i][0]=='$')
+          if(func->used->s1[i][0]=='$' && !func->used->s1[i][1])
              printf("Used in %s\n",func->used->s2[i]);
           else
              printf("Used in %s : %s\n",func->used->s1[i],func->used->s2[i]);
@@ -621,3 +635,64 @@ void WriteWarnRawAppendix(StringList files,StringList2 funcs,StringList2 vars,St
        if(option_warn&WARN_XREF)
           printf("Warning Index : No types to index.\n");
 }
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Decide if to copy or skip the next line.
+
+  int CopyOrSkip Returns the number of characters to skip.
+
+  char *string The string that starts the next line.
+
+  char *type The type of file we are outputing.
+
+  int *copy Returns true if we are to copy the line verbatim.
+
+  int *skip Returns true if we are to skip the line.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+int CopyOrSkip(char *string,char *type,int *copy,int *skip)
+{
+ char *p=string;
+ int s=0;
+
+ if(*p=='\n')
+    p++;
+ while(*p==' ' || *p=='\t')
+    p++;
+
+ *copy=*skip=0;
+
+ switch(*type)
+    {
+    case 'h': /* html */
+     if(!strncmp(p,"+html+",s=6) || !strncmp(p,"-latex-",s=7) || !strncmp(p,"-sgml-",s=6) || !strncmp(p,"-rtf-",s=5))
+        *copy=1;
+     if(!strncmp(p,"-html-",s=6) || !strncmp(p,"+latex+",s=7) || !strncmp(p,"+sgml+",s=6) || !strncmp(p,"+rtf+",s=5) || !strncmp(p,"+none+",s=6))
+        *skip=1;
+     break;
+    case 'l': /* latex */
+     if(!strncmp(p,"-html-",s=6) || !strncmp(p,"+latex+",s=7) || !strncmp(p,"-sgml-",s=6) || !strncmp(p,"-rtf-",s=5))
+        *copy=1;
+     if(!strncmp(p,"+html+",s=6) || !strncmp(p,"-latex-",s=7) || !strncmp(p,"+sgml+",s=6) || !strncmp(p,"+rtf+",s=5) || !strncmp(p,"+none+",s=6))
+        *skip=1;
+     break;
+    case 's': /* sgml */
+     if(!strncmp(p,"-html-",s=6) || !strncmp(p,"-latex-",s=7) || !strncmp(p,"+sgml+",s=6) || !strncmp(p,"-rtf-",s=5))
+        *copy=1;
+     if(!strncmp(p,"+html+",s=6) || !strncmp(p,"+latex+",s=7) || !strncmp(p,"-sgml-",s=6) || !strncmp(p,"+rtf+",s=5) || !strncmp(p,"+none+",s=6))
+        *skip=1;
+     break;
+    case 'r': /* rtf */
+     if(!strncmp(p,"-html-",s=6) || !strncmp(p,"-latex-",s=7) || !strncmp(p,"-sgml-",s=6) || !strncmp(p,"+rtf+",s=5))
+        *copy=1;
+     if(!strncmp(p,"+html+",s=6) || !strncmp(p,"+latex+",s=7) || !strncmp(p,"+sgml+",s=6) || !strncmp(p,"-rtf-",s=5) || !strncmp(p,"+none+",s=6))
+        *skip=1;
+     break;
+    }
+
+ if(*copy)
+    return(p-string+s);
+ else
+    return(0);
+}
