@@ -1,13 +1,13 @@
 /***************************************
-  $Header: /home/amb/cxref/RCS/xref.c 1.15 1997/05/18 10:05:52 amb Exp $
+  $Header: /home/amb/cxref/RCS/xref.c 1.19 1999/06/25 18:09:38 amb Exp $
 
-  C Cross Referencing & Documentation tool. Version 1.4.
+  C Cross Referencing & Documentation tool. Version 1.5b.
 
   Cross referencing of functions.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1995,96,97 Andrew M. Bishop
+  This file Copyright 1995,96,97,99 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -33,6 +33,17 @@
 #include <stdio.h>
 #include <string.h>
 #include <unistd.h>
+#include <limits.h>
+
+#define TYPE_MAX_LEN 256        /*+ The maximum type definition length +*/
+#define FUNC_MAX_LEN 64         /*+ The maximum function name length. +*/
+#if defined(PATH_MAX) && defined(NAME_MAX)
+#define FILE_MAX_LEN (PATH_MAX+NAME_MAX) /*+ The maximum filename length. +*/
+#elif defined(PATH_MAX)
+#define FILE_MAX_LEN (PATH_MAX+256) /*+ The maximum filename length. +*/
+#else
+#define FILE_MAX_LEN 512        /*+ The maximum filename length. +*/
+#endif
 
 #include "memory.h"
 #include "datatype.h"
@@ -61,9 +72,11 @@ static void fixup_extern_var(Variable var,StringList2 refs);
   with the global functions, variables and includes. The types that are defined are also listed here.
 
   File file The file structure containing the information.
+
+  int outputs Set to true if any cross referencing to produce outputs is required.
   ++++++++++++++++++++++++++++++++++++++*/
 
-void CrossReference(File file)
+void CrossReference(File file,int outputs)
 {
  FILE *in,*out;
  char *ifile,*ofile;
@@ -91,7 +104,7 @@ void CrossReference(File file)
 
     if(in)
       {
-       char include[128],filename[128],ch;
+       char include[FILE_MAX_LEN+1],filename[FILE_MAX_LEN+1],ch;
 
        while(fscanf(in,"%s%c",filename,&ch)==2)
          {
@@ -107,8 +120,9 @@ void CrossReference(File file)
              if(diff_file)
                 fprintf(out," %s",include);
 
-             if(include[0]=='%' && !strcmp(&include[1],file->name))
-                AddToStringList(file->inc_in,filename,1,1);
+             if(outputs)
+                if(include[0]=='%' && !strcmp(&include[1],file->name))
+                   AddToStringList(file->inc_in,filename,1,1);
             }
 
           if(diff_file)
@@ -191,7 +205,7 @@ void CrossReference(File file)
 
     if(in)
       {
-       char ch,funcname[64],filename[128],called[64];
+       char ch,funcname[FUNC_MAX_LEN+1],filename[FILE_MAX_LEN+1],called[FUNC_MAX_LEN+1];
        int scope;
 
        while(fscanf(in,"%s %s %d%c",filename,funcname,&scope,&ch)==4)
@@ -200,18 +214,24 @@ void CrossReference(File file)
 
           if(diff_file)
             {
-             if(funcname[0]!='$')
-                check_for_caller(file,funcname,filename);
+             if(outputs)
+                if(funcname[0]!='$' || funcname[1]!=0)
+                   check_for_caller(file,funcname,filename);
              fprintf(out,"%s %s %d",filename,funcname,scope);
             }
 
           while(ch==' ')
             {
              fscanf(in,"%s%c",called,&ch);
+
              if(diff_file)
                {
-                if(called[0]!='%')
-                   check_for_called(file,called,funcname[0]=='$'?NULL:funcname,filename);
+                if(outputs)
+                   if(called[0]!='%')
+                      if(funcname[0]!='$' || funcname[1]!=0)
+                         check_for_called(file,called,funcname,filename);
+                      else
+                         check_for_called(file,called,NULL,filename);
                 fprintf(out," %s",called);
                }
             }
@@ -259,7 +279,7 @@ void CrossReference(File file)
 
     if(in)
       {
-       char varname[64],filename[128],funcname[64],ch;
+       char varname[FUNC_MAX_LEN+1],filename[FILE_MAX_LEN+1],funcname[FUNC_MAX_LEN+1],ch;
        int scope;
 
        while(fscanf(in,"%s %s %d%c",filename,varname,&scope,&ch)==4)
@@ -268,8 +288,9 @@ void CrossReference(File file)
 
           if(diff_file)
             {
-             if(!(scope&LOCAL))
-                check_for_var(file,varname,filename,scope,NULL);
+             if(outputs)
+                if(!(scope&LOCAL))
+                   check_for_var(file,varname,filename,scope,NULL);
              fprintf(out,"%s %s %d",filename,varname,scope);
             }
 
@@ -279,11 +300,12 @@ void CrossReference(File file)
 
              if(diff_file)
                {
-                if(!(scope&LOCAL))
-                   if(funcname[0]=='%')
-                      check_for_var(file,varname,filename,scope,&funcname[1]);
-                   else
-                      check_for_var(file,varname,filename,scope,funcname);
+                if(outputs)
+                   if(!(scope&LOCAL))
+                      if(funcname[0]=='%')
+                         check_for_var(file,varname,filename,scope,&funcname[1]);
+                      else
+                         check_for_var(file,varname,filename,scope,funcname);
                 fprintf(out," %s",funcname);
                }
             }
@@ -298,9 +320,12 @@ void CrossReference(File file)
 
     /* We must fix the location of the extern variables now since it was not known earlier. */
 
-    fixup_extern_var(file->variables,file->v_refs);
-    for(func=file->functions;func;func=func->next)
-       fixup_extern_var(file->variables,func->v_refs);
+    if(outputs)
+      {
+       fixup_extern_var(file->variables,file->v_refs);
+       for(func=file->functions;func;func=func->next)
+          fixup_extern_var(file->variables,func->v_refs);
+      }
 
     fclose(out);
     rename(ofile,ifile);
@@ -330,7 +355,7 @@ void CrossReference(File file)
 
     if(in)
       {
-       char typename[128],filename[128];
+       char typename[TYPE_MAX_LEN+1],filename[FILE_MAX_LEN+1];
 
        while(fscanf(in,"%s %s",filename,typename)==2)
          {
@@ -339,7 +364,7 @@ void CrossReference(File file)
           if(diff_file)
              fprintf(out,"%s %s",filename,typename);
 
-          fgets(typename,128,in);
+          fgets(typename,TYPE_MAX_LEN,in);
 
           if(diff_file)
              fputs(typename,out);
@@ -449,19 +474,28 @@ static void check_for_var(File file,char* variable,char* filename,int scope,char
 {
  Variable var;
 
- for(var=file->variables;var;var=var->next)
-    if(!strcmp(variable,var->name))
-      {
-       if(!funcname)
-         {
-          if(scope&GLOBAL && var->scope&(EXTERNAL|EXTERN_H|EXTERN_F))
-             var->defined=MallocString(filename);
+ if(!funcname)
+   {
+    if(!(scope&(GLOBAL|EXTERNAL|EXTERN_H)))
+       return;
 
-          if(scope&(GLOBAL|EXTERNAL|EXTERN_H) && var->scope&GLOBAL)
-             AddToStringList2(var->visible,"$",filename,1,0);
-         }
-       else
-          if(funcname[0]=='$')
+    for(var=file->variables;var;var=var->next)
+       if((scope&GLOBAL && var->scope&(EXTERNAL|EXTERN_H|EXTERN_F)) ||
+          (scope&(GLOBAL|EXTERNAL|EXTERN_H) && var->scope&GLOBAL))
+          if(!strcmp(variable,var->name))
+            {
+             if(scope&GLOBAL && var->scope&(EXTERNAL|EXTERN_H|EXTERN_F))
+                var->defined=MallocString(filename);
+
+             if(scope&(GLOBAL|EXTERNAL|EXTERN_H) && var->scope&GLOBAL)
+                AddToStringList2(var->visible,"$",filename,1,0);
+            }
+   }
+ else
+   {
+    for(var=file->variables;var;var=var->next)
+       if(!strcmp(variable,var->name))
+          if(funcname[0]=='$' && !funcname[1])
              AddToStringList2(var->used,"$",filename,1,0);
           else
             {
@@ -470,7 +504,7 @@ static void check_for_var(File file,char* variable,char* filename,int scope,char
              if(scope&EXTERN_F && var->scope&GLOBAL)
                 AddToStringList2(var->visible,funcname,filename,1,1);
             }
-      }
+   }
 }
 
 
@@ -543,7 +577,7 @@ static void fixup_extern_var(Variable var,StringList2 refs)
 
  for(i=0;i<refs->n;i++)
    {
-    if(refs->s2[i][0]=='$')
+    if(refs->s2[i][0]=='$' && !refs->s2[i][1])
        for(v=var;v;v=v->next)
           if(v->scope&(EXTERNAL|EXTERN_H|EXTERN_F) && !strcmp(refs->s1[i],v->name))
             {
@@ -591,7 +625,7 @@ void CreateAppendix(StringList files,StringList2 funcs,StringList2 vars,StringLi
 
     if(in)
       {
-       char include[128],filename[128],ch;
+       char include[FILE_MAX_LEN+1],filename[FILE_MAX_LEN+1],ch;
 
        while(fscanf(in,"%s%c",filename,&ch)==2)
          {
@@ -612,7 +646,7 @@ void CreateAppendix(StringList files,StringList2 funcs,StringList2 vars,StringLi
 
     if(in)
       {
-       char ch,caller[64],filename[128],called[64];
+       char ch,caller[FUNC_MAX_LEN+1],filename[FILE_MAX_LEN+1],called[FUNC_MAX_LEN+1];
        int scope;
 
        while(fscanf(in,"%s %s %d%c",filename,caller,&scope,&ch)==4)
@@ -635,7 +669,7 @@ void CreateAppendix(StringList files,StringList2 funcs,StringList2 vars,StringLi
 
     if(in)
       {
-       char variable[64],filename[128],funcname[64],ch;
+       char variable[FUNC_MAX_LEN+1],filename[FILE_MAX_LEN+1],funcname[FUNC_MAX_LEN+1],ch;
        int scope;
 
        while(fscanf(in,"%s %s %d%c",filename,variable,&scope,&ch)==4)
@@ -658,20 +692,20 @@ void CreateAppendix(StringList files,StringList2 funcs,StringList2 vars,StringLi
 
     if(in)
       {
-       char typename[128],filename[128];
+       char typename[TYPE_MAX_LEN],filename[FILE_MAX_LEN+1];
 
        while(fscanf(in,"%s %s",filename,typename)==2)
          {
           if(typename[0]=='#')
             {
-             fgets(typename,128,in);
+             fgets(typename,TYPE_MAX_LEN,in);
              typename[strlen(typename)-1]=0;
              AddToStringList2(types,&typename[1],filename,1,1);
             }
           else
             {
              AddToStringList2(types,typename,filename,1,1);
-             fgets(typename,128,in);
+             fgets(typename,TYPE_MAX_LEN,in);
             }
          }
 
@@ -704,7 +738,7 @@ void CrossReferenceDelete(char *name)
    {fprintf(stderr,"cxref: Failed to open the include cross reference file '%s'\n",ofile);fclose(in);}
  else if(in)
    {
-    char include[128],filename[128],ch;
+    char include[FILE_MAX_LEN+1],filename[FILE_MAX_LEN+1],ch;
 
     while(fscanf(in,"%s%c",filename,&ch)==2)
       {
@@ -749,7 +783,7 @@ void CrossReferenceDelete(char *name)
    {fprintf(stderr,"cxref: Failed to open the functional cross reference file '%s'\n",ofile);fclose(in);}
  else if(in)
    {
-    char ch,funcname[64],filename[128],called[64];
+    char ch,funcname[FUNC_MAX_LEN+1],filename[FILE_MAX_LEN+1],called[FUNC_MAX_LEN+1];
     int scope;
 
     while(fscanf(in,"%s %s %d%c",filename,funcname,&scope,&ch)==4)
@@ -794,7 +828,7 @@ void CrossReferenceDelete(char *name)
    {fprintf(stderr,"cxref: Failed to open the variable cross reference file '%s'\n",ofile);fclose(in);}
  else if(in)
    {
-    char varname[64],filename[128],funcname[64],ch;
+    char varname[FUNC_MAX_LEN+1],filename[FILE_MAX_LEN+1],funcname[FUNC_MAX_LEN+1],ch;
     int scope;
 
     while(fscanf(in,"%s %s %d%c",filename,varname,&scope,&ch)==4)
@@ -840,7 +874,7 @@ void CrossReferenceDelete(char *name)
    {fprintf(stderr,"cxref: Failed to open the typedef reference file '%s'\n",ofile);fclose(in);}
  else if(in)
    {
-    char typename[128],filename[128];
+    char typename[TYPE_MAX_LEN+1],filename[FILE_MAX_LEN+1];
 
     while(fscanf(in,"%s %s",filename,typename)==2)
       {
@@ -849,7 +883,7 @@ void CrossReferenceDelete(char *name)
        if(diff_file)
           fprintf(out,"%s %s",filename,typename);
 
-       fgets(typename,128,in);
+       fgets(typename,TYPE_MAX_LEN,in);
 
        if(diff_file)
           fputs(typename,out);
